@@ -11,8 +11,9 @@ import base64
 import binascii
 import datetime
 from collections import namedtuple
-from urllib.parse import urlparse
-
+from urllib.parse import urlencode, urlparse
+from tornado.httpclient import AsyncHTTPClient, HTTPRequest
+from tornado.httputil import url_concat
 from tornado.httpserver import HTTPServer
 from tornado.ioloop import IOLoop
 from tornado.web import (
@@ -22,9 +23,14 @@ from tornado.web import (
     Finish,
     MissingArgumentError,
 )
-from jupyterhub.services.auth import HubAuthenticated
+from jupyterhub.services.auth import (
+    HubOAuthenticated,
+)
 from sqlalchemy import create_engine, or_
 from sqlalchemy.orm import sessionmaker
+from tornado import log
+import requests
+
 
 try:
     from . import dbutil
@@ -61,6 +67,58 @@ except ImportError:  # pragma: no cover
     )
 
 
+class JupyterHubLoginHandler(RequestHandler):
+    """
+    Login Handler
+    This handler both begins and ends the OAuth process
+    """
+
+    async def token_for_code(self, code):
+        """
+        Complete OAuth by requesting an access token for an oauth code
+        """
+        params = dict(
+            client_id=self.settings['client_id'],
+            client_secret=self.settings['api_token'],
+            grant_type='authorization_code',
+            code=code,
+            redirect_uri=self.settings['redirect_uri'],
+        )
+        req = HTTPRequest(
+            self.settings['token_url'],
+            method='POST',
+            body=urlencode(params).encode('utf8'),
+            headers={'Content-Type': 'application/x-www-form-urlencoded'},
+        )
+        response = await AsyncHTTPClient().fetch(req)
+        data = json.loads(response.body.decode('utf8', 'replace'))
+        return data['access_token']
+
+    async def get(self):
+        code = self.get_argument('code', None)
+        if code:
+            # code is set, we are the oauth callback
+            # complete oauth
+            token = await self.token_for_code(code)
+            # login successful, set cookie and redirect back to home
+            self.set_secure_cookie('ngshare-oauth-token', token)
+            self.redirect('/services/ngshare/')
+        else:
+            # we are the login handler,
+            # begin oauth process which will come back later with an
+            # authorization_code
+            self.redirect(
+                url_concat(
+                    self.settings['authorize_url'],
+                    dict(
+                        redirect_uri=self.settings['redirect_uri'],
+                        client_id=self.settings['client_id'],
+                        response_type='code',
+                    ),
+                )
+            )
+
+
 class MyHelpers:
     'Helper functions for database accesses'
 
@@ -85,17 +143,17 @@ class MyHelpers:
                 self.json_error(400, 'Time format incorrect')
 
     def path_check(self, pathname):
-        '''
-            Return whether a pathname (for file, in director tree) is safe
-            Current policy:
-                Not empty
-                os.path.abspath resolves to a child address
-                Path does not contain ('.', '..', '', '/')
-            Note: os.path.abspath is used instead of os.path.realpath to prevent
-             symbolic link issues, because the file is not on server
-            Note: os.path.abspath is not 100% safe
-            Note: currently only using Linux pathname conventions
-        '''
+        """
+        Return whether a pathname (for file, in director tree) is safe
+        Current policy:
+            Not empty
+            os.path.abspath resolves to a child address
+            Path does not contain ('.', '..', '', '/')
+        Note: os.path.abspath is used instead of os.path.realpath to prevent
+         symbolic link issues, because the file is not on server
+        Note: os.path.abspath is not 100% safe
+        Note: currently only using Linux pathname conventions
+        """
         if not pathname:
             return False
         path = pathname
@@ -133,11 +191,11 @@ class MyHelpers:
         return ans
 
     def json_files_unpack(self, json_str, target):
-        '''
-            Generate a list of File objects from a JSON directory tree
-            json_str: json object as string; raise error when None
-            target: a list to put file objects in
-        '''
+        """
+        Generate a list of File objects from a JSON directory tree
+        json_str: json object as string; raise error when None
+        target: a list to put file objects in
+        """
         if json_str is None:
             self.json_error(400, 'Please supply files')
         try:
@@ -350,7 +408,7 @@ class MyHelpers:
             self.json_error(403, msg)
 
 
-class MyRequestHandler(HubAuthenticated, RequestHandler, MyHelpers):
+class MyRequestHandler(HubOAuthenticated, RequestHandler, MyHelpers):
     'Custom request handler for ngshare'
 
     def json_success(self, msg=None, **kwargs):
@@ -368,10 +426,14 @@ class MyRequestHandler(HubAuthenticated, RequestHandler, MyHelpers):
         raise Finish(json.dumps({'success': False, 'message': msg, **kwargs}))
 
     def prepare(self):
-        'Provide a db object'
+        'Runs before web requests, initializes user and db session'
         self.db = self.application.db_session()
+
+        self.auth_token = self.get_current_token()
+
         current_user = self.get_current_user()
-        if current_user is not None:
+
+        if current_user is not None and "name" in current_user.keys():
             self.user = User.from_jupyterhub_user(current_user, self.db)
         else:
             self.user = None
@@ -379,6 +441,35 @@ class MyRequestHandler(HubAuthenticated, RequestHandler, MyHelpers):
     def on_finish(self):
         self.db.close()
 
+    def get_current_user(self):
+        """
+        @web.authenticated calls this method.
+        If a Falsy value is returned, the request is redirected to `login_url`.
+        If a Truthy value is returned, the request is allowed to proceed.
+        """
+        if self.auth_token:
+            return self.user_for_token(self.auth_token)
+
+    def get_current_token(self):
+        'Gets token from either the headers, if provided, or from a cookie that should have been set earlier.'
+        if "Authorization" in self.request.headers:
+            return self.request.headers.get('Authorization')[6:]
+        else:
+            # The login handler stored a JupyterHub API token in a cookie
+            token = self.get_secure_cookie('ngshare-oauth-token')
+            if token:
+                # secure cookies are bytes, decode to str
+                return token.decode('ascii', 'replace')
+
+    def user_for_token(self, token):
+        'Retrieve the user for a given token, via /hub/api/user'
+        r = requests.get(
+            self.settings['user_url'],
+            headers={'Authorization': f'token {token}'},
+        )
+        if r.ok:
+            return r.json()
+
 
 class HomePage(MyRequestHandler):
     '/api/'
@@ -409,10 +500,10 @@ class ListCourses(MyRequestHandler):
 
     @authenticated
     def get(self):
-        '''
-            List all available courses the user is taking or teaching. (anyone)
-            List all courses in ngshare. (admin)
-        '''
+        """
+        List all available courses the user is taking or teaching. (anyone)
+        List all courses in ngshare. (admin)
+        """
         courses = set()
         if self.is_admin():
             for i in self.db.query(Course).all():
@@ -460,10 +551,10 @@ class ManageInstructor(MyRequestHandler):
 
     @authenticated
     def post(self, course_id, instructor_id):
-        '''
-            Add or update a course instructor. (admin)
-            Update self full name or email. (instructors)
-        '''
+        """
+        Add or update a course instructor. (admin)
+        Update self full name or email. (instructors)
+        """
         course = self.find_course(course_id)
         self.check_course_instructor(course)
         instructor = self.find_or_create_user(instructor_id)
@@ -564,10 +655,10 @@ class ManageStudent(MyRequestHandler):
 
     @authenticated
     def get(self, course_id, student_id):
-        '''
-            Get information about a student.
-            (instructors+student with same student_id)
-        '''
+        """
+        Get information about a student.
+        (instructors+student with same student_id)
+        """
         course = self.find_course(course_id)
         if self.user.id != student_id:
             self.check_course_instructor(course)
@@ -635,7 +726,10 @@ class ListStudents(MyRequestHandler):
             association.last_name = i['last_name']
             association.email = i['email']
             ans.append(
-                {'username': i['username'], 'success': True,}
+                {
+                    'username': i['username'],
+                    'success': True,
+                }
             )
         self.db.commit()
         self.json_success(status=ans)
@@ -705,10 +799,10 @@ class ListSubmissions(MyRequestHandler):
     '/api/submissions/<course_id>/<assignment_id>'
 
     def get(self, course_id, assignment_id):
-        '''
-            List all submissions for an assignment from all students
-             (instructors only)
-        '''
+        """
+        List all submissions for an assignment from all students
+         (instructors only)
+        """
         course = self.find_course(course_id)
         self.check_course_instructor(course)
         assignment = self.find_assignment(course, assignment_id)
@@ -727,11 +821,11 @@ class ListStudentSubmissions(MyRequestHandler):
     '/api/submissions/<course_id>/<assignment_id>/<student_id>'
 
     def get(self, course_id, assignment_id, student_id):
-        '''
-            List all submissions for an assignment from a particular student
-             (instructors+students,
-              students restricted to their own submissions)
-        '''
+        """
+        List all submissions for an assignment from a particular student
+         (instructors+students,
+          students restricted to their own submissions)
+        """
         course = self.find_course(course_id)
         if self.user.id != student_id:
             self.check_course_instructor(course)
@@ -767,10 +861,10 @@ class DownloadAssignment(MyRequestHandler):
     '/api/submission/<course_id>/<assignment_id>/<student_id>'
 
     def get(self, course_id, assignment_id, student_id):
-        '''
-            Download a student's submitted assignment (instructors only)
-            TODO: maybe allow student to see their own submissions?
-        '''
+        """
+        Download a student's submitted assignment (instructors only)
+        TODO: maybe allow student to see their own submissions?
+        """
         course = self.find_course(course_id)
         self.check_course_instructor(course)
         assignment = self.find_assignment(course, assignment_id)
@@ -795,10 +889,10 @@ class UploadDownloadFeedback(MyRequestHandler):
     '/api/feedback/<course_id>/<assignment_id>/<student_id>'
 
     def post(self, course_id, assignment_id, student_id):
-        '''
-            POST /api/feedback/<course_id>/<assignment_id>/<student_id>
-            Upload feedback on a student's assignment (instructors only)
-        '''
+        """
+        POST /api/feedback/<course_id>/<assignment_id>/<student_id>
+        Upload feedback on a student's assignment (instructors only)
+        """
         course = self.find_course(course_id)
         self.check_course_instructor(course)
         assignment = self.find_assignment(course, assignment_id)
@@ -819,11 +913,11 @@ class UploadDownloadFeedback(MyRequestHandler):
         self.json_success()
 
     def get(self, course_id, assignment_id, student_id):
-        '''
-            GET /api/feedback/<course_id>/<assignment_id>/<student_id>
-            Download feedback on a student's assignment
-             (instructors+students, students restricted to own submissions)
-        '''
+        """
+        GET /api/feedback/<course_id>/<assignment_id>/<student_id>
+        Download feedback on a student's assignment
+         (instructors+students, students restricted to own submissions)
+        """
         course = self.find_course(course_id)
         if self.user.id != student_id:
             self.check_course_instructor(course)
@@ -871,7 +965,11 @@ class InitDatabase(MyRequestHandler):
                 for line in value:
                     tbody.append(list(map(line.__getitem__, thead)))
                 ans.append(
-                    {'header': key, 'thead': thead, 'tbody': tbody,}
+                    {
+                        'header': key,
+                        'thead': thead,
+                        'tbody': tbody,
+                    }
                 )
             self.render('dump.html', tables=ans)
         else:
@@ -936,11 +1034,28 @@ class MyApplication(Application):
                 UploadDownloadFeedback,
             ),
             (prefix + 'initialize-Data6ase', InitDatabase),
+            (prefix + 'oauth_callback', JupyterHubLoginHandler),
+            (prefix + 'healthz', HealthCheckHandler),
             ('/healthz', HealthCheckHandler),
         ]
+        hub_api = os.environ['JUPYTERHUB_API_URL'].rstrip('/')
+        authorize_url = hub_api + '/oauth2/authorize'
+        token_url = hub_api + '/oauth2/token'
+        user_url = hub_api + '/user'
         handlers.append((r'.*', NotFoundHandler))
         super(MyApplication, self).__init__(
-            handlers, debug=debug, autoreload=autoreload
+            handlers,
+            debug=debug,
+            autoreload=autoreload,
+            cookie_secret=os.urandom(32),
+            login_url='/oauth_callback',
+            api_token=os.environ['JUPYTERHUB_API_TOKEN'],
+            client_id=os.environ['JUPYTERHUB_CLIENT_ID'],
+            redirect_uri=os.environ['JUPYTERHUB_SERVICE_PREFIX'].rstrip('/')
+            + '/oauth_callback',
+            authorize_url=authorize_url,
+            token_url=token_url,
+            user_url=user_url,
         )
         # Connect Database
         engine = create_engine(db_url)
@@ -953,15 +1068,18 @@ class MyApplication(Application):
         self.admin = admin
 
 
-class MockAuth(HubAuthenticated):
-    '''
-        Mock class substituting HubAuthenticated, for vngshare
-    '''
+class MockAuth:
+    """
+    Mock class containing methods to replace request handler methods.
+    """
 
     def get_login_url(self):
         return 'http://example.com/'
 
-    def get_current_user(self):
+    def get_current_token(self):
+        return "token"
+
+    def user_for_token(self, _):
         if type(self).__name__ in ('HomePage', 'Static', 'InitDatabase'):
             user = self.get_argument('user', 'user')
         else:
